/**	$Revision: 1.29 $
	Last updated: $Date: 2006/03/06 21:00:35 $

	Copyright (C) 2000-2002 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include "FreeCoins.h"
#include "CoinsPref.h"
#include "AccountsDB.h" 
#include "TransactionsDB.h" 
#include "SplitsDB.h"
#include "CurrencyDB.h"
#include "FormManager.h"
#include "AccountsForm.h"
#include "FreeCoinsRsc.h"
#include "GenericInputForm.h"

TransactionsDB trsdb;
AccountsDB accdb;
SplitsDB spldb;
CurrencyDB curdb;

Boolean ApplicationHandleEvent( EventPtr event )
{
	Boolean	eventHandled = true;
	UInt16 formID;
	FormPtr frm;
	
	switch (event->eType) {
		case frmLoadEvent:
			// Load the form resource.		
			formID = event->data.frmLoad.formID;                
			if(FormManager::instance()->dispatchLoadEvent(formID)) {
				return true;
			}
			break;

		case frmCloseEvent:
			//first try our form manager
			if(FormManager::instance()->dispatchCloseEvent(event->data.frmClose.formID)) {
				return true;
			}
			//if that failed carry on with the default close operation
			frm = FrmGetFormPtr(event->data.frmClose.formID);
			FrmEraseForm(frm);
			FrmDeleteForm(frm);
			eventHandled = true;
			break;

		default:
			eventHandled = false;
	}	
	return eventHandled;
}


void ApplicationEventLoop(void)
{
	UInt16 error;
	EventType event;

	do {
		EvtGetEvent(&event, evtWaitForever);
		if(!SysHandleEvent(&event)) //not system event
			if(!MenuHandleEvent(NULL,&event,&error))
				//our event handling function
				if(!ApplicationHandleEvent(&event))
					if(!FormManager::instance()->dispatchEvent(&event)) 
						FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

static void StartApplication( void )
{
	UInt32 datePref;
        
	//init the database pointers
	transactions_db = &trsdb;
	splits_db = &spldb;
	currency_db = &curdb;
	accounts_db = &accdb;

	//set some localisation parameters
	LocGetNumberSeparators((NumberFormatType) PrefGetPreference(prefNumberFormat),
				   &MiscFunc::cThousandSep, &MiscFunc::cDecimalSep);	
	datePref = PrefGetPreference(prefDateFormat);
	switch(datePref) {
		case dfDMYWithSlashes: 
			MiscFunc::bDayFirst = true; MiscFunc::cDateSep='/';
			break;
		case dfDMYWithDots: 
			MiscFunc::bDayFirst = true; MiscFunc::cDateSep='.';
			break;
		case dfDMYWithDashes: 
			MiscFunc::bDayFirst = true; MiscFunc::cDateSep='-';
			break;
		default:
		case dfYMDWithSlashes:
		case dfMDYWithSlashes: 
			MiscFunc::bDayFirst = false; MiscFunc::cDateSep='/';
			break;
		case dfYMDWithDashes: 
			MiscFunc::bDayFirst = false; MiscFunc::cDateSep='-';
			break;              
		case dfYMDWithDots: 
			MiscFunc::bDayFirst = false; MiscFunc::cDateSep='.';
			break;              
	}

	//update scheduled transactions
	transactions_db->schedule();
	accounts_db->updateVisibility();
}


static void StopApplication( void )
{
	// Close all open forms,  this will force any unsaved data to
	// be written to the database. Also need to save the preferences and dispose of allocated memory
	FrmCloseAllForms();
	preferences.save();
	FormManager::instance()->suicide();
}

/* Run application with 'GlobalFind Search' Launch code */
static void RunSearch(FindParamsPtr findParams)
{
	UInt16			cardNo = 0;
	UInt16			fieldNum = 0;		// unused
	UInt16			recordNum;
	Char*			header;
	Char*			match;
	LocalID			dbID;
	MemHandle		h;
	DmOpenRef		dbP;
	RectangleType		r;
	UInt16			matchPos;
	UInt16			matchLength = 0;	// unused
	CoinsPref		pref1;

	if (pref1.get()->check_pass) return;	/* If password-protected, do not use global find */

	// Open the application's data file.
        dbID = DmFindDatabase(cardNo, TRANSACTIONSDB_NAME);
		
	if (!dbID) {
		    findParams->more = false;
		    return;
	}
	// Display the heading line.
	header = "FreeCoins";
	if (FindDrawHeader(findParams, header)) return;

	// Open the database.
	dbP = DmOpenDatabase(cardNo, dbID, findParams->dbAccesMode);
	if (!dbP) {
		findParams->more = false;
		return;
	}
	// Search fields for the "find" string.
	recordNum = findParams->recordNum;

	while (true)
	{
		if ((recordNum & 0x000f) == 0 && // every 16th record give chance to abort find
			EvtSysEventAvail(true))
		{
			findParams->more = true;
			break;
		}

		h = DmQueryNextInCategory (dbP, &recordNum, dmAllCategories);

		//Stop searching if there are no more records.
		if (!h) {
			findParams->more = false;
			break;
		}


		// get the handle to the record and set busy bit
		// DmQueryRecord() does not mark the busy bit of records.
		h = DmGetRecord(dbP, recordNum);
		//Stop searching if out of memory
		if (!h) {
			findParams->more = true;
			break;
		}
		
		packedTransactionStructure * packed_trans = (packedTransactionStructure *) MemHandleLock(h);
		Char *payee = NULL;
		Char *note = NULL;
		Char *num = NULL;

		// decode packed structure to get interesting fields
		Char* s = (Char*)packed_trans + sizeof(packedTransactionStructure);
		if (!packed_trans->is_scheduled) {	// only interested in non-scheduled events
			payee = s;
			s += StrLen(s) + 1;
			num = s;
			s += StrLen(s) + 1;
			note = s;
		}

		if (payee && FindStrInStr(payee, findParams->strToFind, &matchPos))
			match = payee;
		else if (note && FindStrInStr(note, findParams->strToFind, &matchPos))
			match = note;
		else
			match = NULL;
		    

		DmReleaseRecord(dbP, recordNum, true);

		if (match) {
			// Add the match to the find paramter block,  if there is no
			// room to display the match the following function will
			// return true.
			if (FindSaveMatch (findParams, recordNum, matchPos,
				fieldNum, matchLength, cardNo, dbID)) {
				MemHandleUnlock (h);
				break;
			}

			// Get the bounds of the region where we
			// will draw the results.
			FindGetLineBounds (findParams, &r);

			// Display the match (in future, do not hardcode that 150 chars...)
			WinDrawTruncChars (match, StrLen(match), r.topLeft.x+1, r.topLeft.y, 150);	// 
			findParams->lineNumber++;
		}


		MemHandleUnlock (h);
		recordNum++;
	}

	DmCloseDatabase (dbP);
}

/* Run application with 'GlobalFind GoTo' Launch code */
static void RunGoTo(GoToParamsType *gotoParams)
{
	UInt32 trans_uid = transactions_db->getRecordUid(gotoParams->recordNum);
	UInt32 acc_uid = transactions_db->getPeerAcc(trans_uid, 0);
	accounts_db->selectRecord(acc_uid);
	transactions_db->selectRecord(gotoParams->recordNum);
	FrmPopupForm(EDIT_TRANS_FORM);
}

/* Run application with normal Launch code */
static void RunApp(void) 
{
	// check password if desired
	if (preferences.get()->check_pass) {
		// just in case the pass was deactivated after the user set the
		// option in the prefs dialog
		MemHandle str_handle;
		if (!PwdExists()) {
			alert_res (GENERIC_INFO, PASS_ERR_STR);
		} else {
			MemHandle msg_handle;
			while(1) {
				Char pwd_str[pwdLength];
				str_handle = DmGetResource (strRsc,PASS_TITLE_STR);
				msg_handle = DmGetResource (strRsc,PASS_MSG_STR);
				GenericInputForm g(reinterpret_cast<char*> (MemHandleLock (str_handle)),
					reinterpret_cast<char*> (MemHandleLock (msg_handle)),pwd_str,pwdLength);
				if(g.doDialog()) {
					if(PwdVerify(pwd_str)) break;
				} else {
					unlock_release (msg_handle);
					unlock_release (str_handle);
					StopApplication();
					return;
				}
			}
			unlock_release (msg_handle);
			unlock_release (str_handle);
		}
	}
	//we can start the application
	StartApplication();
	
	//go straight to the accounts form on startup
	AccountsForm accounts_form;
	accounts_form.popup();

	// run the event loop. returns to exit.		
	ApplicationEventLoop();
	StopApplication();
}

/*	Application entry point  */
UInt32	PilotMain( UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags )
{
	if(!osVerGe(palmos_3_5)) {
		FrmCustomAlert(GENERIC_INFO,"FreeCoins will not run on PalmOS versions prior to 3.5",NULL,NULL);
		FrmCloseAllForms();
		return 1;
	}

	switch (cmd) {
		case sysAppLaunchCmdNormalLaunch:	// If normal launch
			RunApp();
			break;
			
		case sysAppLaunchCmdSaveData:		// Find Button pressed while application, active, save data
			//FrmSaveAllForms();
			break;
			
		case sysAppLaunchCmdFind:		// If GlobalFind search
			RunSearch((FindParamsPtr)cmdPBP);
                        break;

		case sysAppLaunchCmdGoTo:		// if GlobalFind goto
			if (launchFlags & sysAppLaunchFlagNewGlobals) {	// we are not yet running, init...
				StartApplication();

				//go straight to the accounts form on startup
				AccountsForm accounts_form;
				accounts_form.popup();

				RunGoTo((GoToParamsType *) cmdPBP);

				// run the event loop. returns to exit.	
				ApplicationEventLoop();
				StopApplication();
			} else {	// we were already running before Find, no need to reinit
				RunGoTo((GoToParamsType *) cmdPBP);
			}
			break;
        }
	return 0;
}	
