/**	$Revision: 1.18 $
	Last updated: $Date: 2006/02/26 14:19:11 $
	
	Copyright (C) 2001-2002 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <PalmOS.h>

#include "SplitsDB.h"
#include "AccountsDB.h"
#include "TransactionsDB.h"
#include "MiscFunc.h" //macro defs
#include "FreeCoinsRsc.h"

SplitsDB* splits_db;

void SplitsDB::unpackRecord (splitStructure* spl, packedSplitStructure* packed_spl)
{
	spl->trans_uid = packed_spl->trans_uid;
	spl->acc_uid = packed_spl->acc_uid;
	spl->amount = packed_spl->amount;
	spl->line_no = packed_spl->line_no;
}

void SplitsDB::packRecord (splitStructure* spl, MemHandle db_entry)
{
	UInt32	length = 0;
	MemPtr	s;
	UInt32	offset = 0;

	length = sizeof(packedSplitStructure);

	// resize the record if necessary and write the data
	if (MemHandleResize(db_entry, length) == 0) {
		s = MemHandleLock(db_entry);
		writeAndAdvance(s, offset, &spl->trans_uid);
		writeAndAdvance(s, offset, &spl->acc_uid);
		writeAndAdvance(s, offset, &spl->amount);
		writeAndAdvance(s, offset, &spl->line_no);
		MemHandleUnlock(db_entry);
	} else {
		ErrFatalDisplay ("resizing failed");
	}
}

SplitsDB::SplitsDB() : Database <splitStructure, packedSplitStructure> (
	SPLITSDB_TYPE, FREE_COINS_CREATOR, SPLITSDB_NAME, SPLITS_CATEGORIES)
{
	current_cat=dmAllCategories;
}

UInt16 SplitsDB::locateFirstSplit(UInt32 trans_uid)
{
	packedSplitStructure s;
	UInt16 recordNo;

	// Nifty trick: I just use DmFindSortPosition to see where an entry
	// with line number 0 would be inserted, which uses binary search
	s.trans_uid = trans_uid;
	s.acc_uid = 0;
	s.amount = 0;
	s.line_no = 0;

	recordNo = DmFindSortPosition(ref, &s, NULL, (DmComparF*)compareRecords, 0);

//#define TRY_TO_FIX_NIFTY_TRICK
#ifdef TRY_TO_FIX_NIFTY_TRICK	
	if (recordNo == 0) {	// WTF, "Nifty trick" sometimes fails and returns 0 ?? Why ?!
		splitStructure sp;
		/* code below always works, but is SLOW, so we try it only if we detect DmFindSortPosition failed -- /mn/ */
		for (recordNo=0; DmQueryNextInCategory(splits_db->getRef(), &recordNo, dmAllCategories); recordNo++) {
			splits_db->getUnpackedRecord(recordNo, &sp);
			if (sp.trans_uid == trans_uid) {
				return recordNo;
			}
		}
	}
#endif

//#define DOUBLECHECK_NIFTY_TRICK
#ifdef DOUBLECHECK_NIFTY_TRICK
	// doublecheck that we got something reasonable.
	if (recordNo > 0) {
		splitStructure sp;
		splits_db->getUnpackedRecord(recordNo, &sp);	// needed only for ErrFatalDisplayIf
		ErrFatalDisplayIf(sp.trans_uid != trans_uid, "bug in locateFirstSplit, use ForceResort");
	}
#endif	

	return recordNo;
}

void SplitsDB::clearRecord(UInt16 id)
{
	MemHandle h;
	splitStructure s;

	UInt16 cleared_cat = findcat_res (CLEARED_STR);
	if (getRecordCategory(id)==cleared_cat)
		return;
	setRecordCategory(id,cleared_cat); 
	h = getUnpackedRecord(id, &s);
	accounts_db->incrementBalance(s.acc_uid, s.amount, 0);
}

void SplitsDB::unclearRecord(UInt16 id)
{
	MemHandle h;
	splitStructure s;

	UInt16 uncleared_cat = findcat_res (UNCLEARED_STR);
	if (getRecordCategory(id)==uncleared_cat)
		return;
	setRecordCategory(id,uncleared_cat); 
	h = getUnpackedRecord(id, &s);
	accounts_db->incrementBalance(s.acc_uid, -s.amount, 0);
}

UInt16	SplitsDB::newRecord(splitStructure* s, Boolean cleared)
{
	MemHandle					h,p;
	UInt16						index=0;

	h=DmNewHandle(ref,1);
	ErrFatalDisplayIf(!ref,"Ref is 0");

	packRecord(s,h);	

	if(!h) {
		fatal_res (NEW_SPLIT_FAILED_STR);
	}
	p= (MemHandle) MemHandleLock(h);
	index=DmFindSortPosition(ref,p, 0, (DmComparF *) compareRecords, 0);
	MemHandleUnlock(h);
	
	DmAttachRecord(ref,&index,h,0);	

	selectRecord(index);
	if(!transactions_db->isScheduled(transactions_db->getRecordId(s->trans_uid))) { //if the transaction is not scheduled
		//update the balance of the involved account
		accounts_db->incrementBalance(s->acc_uid, cleared?s->amount:0, s->amount);
	}
	
	//set the category
	MemHandle str_handle = DmGetResource (strRsc,cleared?CLEARED_STR:UNCLEARED_STR);;
	setRecordCategory(index,CategoryFind(ref,reinterpret_cast<char*> (MemHandleLock (str_handle))));
	unlock_release (str_handle);

	return index;
}

bool SplitsDB::deleteRecord(UInt16 id, Boolean keep_balance)
{
	splitStructure		s;
	Boolean 			cleared;

	getUnpackedRecord(id,&s);

	MemHandle str_handle = DmGetResource (strRsc,CLEARED_STR);
	cleared = (getRecordCategory(id)==CategoryFind(ref,reinterpret_cast<char*> (MemHandleLock (str_handle))));
	unlock_release (str_handle);
				
	//update the balance of the involved account
	if(!keep_balance) 
		accounts_db->incrementBalance(s.acc_uid, cleared?-s.amount:0, -s.amount);

	return Database<splitStructure, packedSplitStructure>::deleteRecord(id);	
}

Int16	SplitsDB::compareRecords(
	packedSplitStructure* rec1, packedSplitStructure* rec2, Int16 other,
	SortRecordInfoPtr sr1, SortRecordInfoPtr sr2, MemHandle app_info)
{
	if (rec1->trans_uid == rec2->trans_uid)
		return rec1->line_no - rec2->line_no;

	return (rec1->trans_uid > rec2->trans_uid) ? 1 : -1;
}
